"""Test PsqlParser."""

from ..psqlparser import PsqlParser


def test_output_has_new_prompt() -> None:
    p = PsqlParser()

    case_trivial = "\x1b[?2004hpgdb=> "
    assert p.output_has_new_prompt(case_trivial)

    case_mixed_results = "\x1b[?2004l\r order_id | order_total_eur | customer_id \r\n----------+-----------------+-------------\r\n(0 rows)\r\n\r\n\x1b[?2004hpgdb=# "
    assert p.output_has_new_prompt(case_mixed_results)


def test_parse_new_prompt_and_rest() -> None:
    p = PsqlParser()

    case_trivial = "\x1b[?2004hpgdb=> "
    assert p.parse_new_prompt_and_rest(case_trivial) == ["", "\x1b[?2004hpgdb=> "]

    case_mixed_results = "\x1b[?2004l\r order_id | order_total_eur | customer_id \r\n----------+-----------------+-------------\r\n(0 rows)\r\n\r\n\x1b[?2004hpgdb=# "
    assert p.parse_new_prompt_and_rest(case_mixed_results) == [
        "\x1b[?2004l\r order_id | order_total_eur | customer_id \r\n----------+-----------------+-------------\r\n(0 rows)\r\n\r\n",
        "\x1b[?2004hpgdb=# ",
    ]


def test_output_has_magical_return() -> None:
    p = PsqlParser()

    case_long_return = "\r\x1b[16Ppgdb=# SELECT * FROM orders  WHERE order_total_eur = 0 AND order_tot\x08\r\n\x1b[?2004l\r"
    assert p.output_has_magical_return(case_long_return)


def test_parse_last_stmt() -> None:
    p = PsqlParser()

    case_trivial = 'psql (14.5)\nType "help" for help.\n\npgdb=# SELECT * FROM orders;'
    assert p.parse_last_stmt(case_trivial) == "SELECT * FROM orders;"

    case_select_then_select = 'psql (14.5)\nType "help" for help.\n\npgdb=# SELECT * FROM orders;\npgdb=# SELECT * FROM orders WHERE order_total_eur = 0 AND order_total_eur = 100;'
    assert (
        p.parse_last_stmt(case_select_then_select)
        == "SELECT * FROM orders WHERE order_total_eur = 0 AND order_total_eur = 100;"
    )

    # Broken case found in empirical testing
    case_empirical_1 = 'sql (14.5)\nType "help" for help.\n\npgdb=# SELECT * FROM orders  WHERE order_total_eur = 0 AND order_total_eur = 100;\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n'
    assert (
        p.parse_last_stmt(case_empirical_1)
        == "SELECT * FROM orders  WHERE order_total_eur = 0 AND order_total_eur = 100;"
    )

    case_query_with_newline = 'sql (14.5)\nType "help" for help.\n\npgdb=# SELECT * FROM orders WHERE\norder_total_eur = 100;\n\n\n'
    assert (
        p.parse_last_stmt(case_query_with_newline)
        == "SELECT * FROM orders WHERE order_total_eur = 100;"
    )
    # current unfixed bug case, needs to be fixed after mid-presentations:
    case_select_then_insert = 'psql (14.5)\nType "help" for help.\n\npgdb=# SELECT * FROM orders;\npgdb=# INSERT INTO orders VALUES (6, 6, 6);'
    assert p.parse_last_stmt(case_select_then_insert) == ""

    case_multiple_queries_and_whitespaces = 'psql (14.5)\nType "help" for help.\n\npgdb=# SELECT * FROM orders;\npgdb=# INSERT INTO orders VALUES (6, 6, 6);\npgdb=#   SELECT    * FROM\n  orders    WHERE order_total_eur = 100   ; '
    assert (
        p.parse_last_stmt(case_multiple_queries_and_whitespaces)
        == "SELECT    * FROM   orders    WHERE order_total_eur = 100   ;"
    )

    case_multiline_query = (
        'psql (14.5)\nType "help" for help.\n\npgdb=# SELECT * FROM\npgdb-# orders;'
    )
    assert p.parse_last_stmt(case_multiline_query) == "SELECT * FROM  orders;"



import copy
from typing import Optional

import sqlglot.expressions as exp
from psycopg import Connection

from .errfmt import ErrorFormatter
from .qepparser import QEPAnalysis, QEPNode, QEPParser
from .sqlparser import Column, SqlParser


class ImpliedExpressionChecker:
    def __init__(
        self, parsed_sql: exp.Expression, sql_statement: str, db_connection: Connection
    ):
        self.parsed_sql: exp.Expression = parsed_sql
        self.sql_statement: str = sql_statement
        self.db_connection: Connection = db_connection

    def check(self) -> Optional[str]:
        """
        Returns warning_msg if implied expression is detected, otherwise None.
        """

        def finder(node: QEPNode) -> bool:
            return node.get("One-Time Filter") != None

        qep_parser_with_constraint_exclusion = QEPParser(
            conn=self.db_connection, constraint_exclusion=True
        )
        qep_analysis_with_constraint_exclusion = (
            qep_parser_with_constraint_exclusion.parse(self.sql_statement)
        )
        if qep_analysis_with_constraint_exclusion is None:
            return None
        has_onetime_filter_with_constraint_exclusion = (
            len(qep_analysis_with_constraint_exclusion.root.rfind(finder)) > 0
        )

        qep_parser_without_constraint_exclusion = QEPParser(
            conn=self.db_connection, constraint_exclusion=False
        )
        qep_analysis_without_constraint_exclusion = (
            qep_parser_without_constraint_exclusion.parse(self.sql_statement)
        )
        if qep_analysis_without_constraint_exclusion is None:
            return None
        has_onetime_filter_without_constraint_exclusion = (
            len(qep_analysis_without_constraint_exclusion.root.rfind(finder)) > 0
        )

        has_implied_expression = (
            has_onetime_filter_with_constraint_exclusion
            and not has_onetime_filter_without_constraint_exclusion
        )

        if not has_implied_expression:
            return None

        warning = "Found impossible comparison due to column/table constraints"
        warning_name = type(self).__name__.rstrip("Checker")

        formatter = ErrorFormatter(warning, warning_name)
        warning_msg = formatter.format()

        return warning_msg
